@@ -28,6 +28,7 @@
 #include "eloop.h"
 #include "accounting.h"
 #include "driver.h"
+#include "ieee802_1x.h"
 
 
 static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
@@ -75,8 +76,7 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 	}
 
 	if (sta) {
-		val = sta->identity;
-		len = sta->identity_len;
+		val = ieee802_1x_get_identity(sta->eapol_sm, &len);
 		if (!val) {
 			snprintf(buf, sizeof(buf), RADIUS_ADDR_FORMAT,
 				 MAC2STR(sta->addr));
@@ -141,6 +141,13 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 			printf("Could not add Connect-Info\n");
 			goto fail;
 		}
+
+		val = ieee802_1x_get_radius_class(sta->eapol_sm, &len);
+		if (val &&
+		    !radius_msg_add_attr(msg, RADIUS_ATTR_CLASS, val, len)) {
+			printf("Could not add Class\n");
+			goto fail;
+		}
 	}
 
 	return msg;
@@ -214,7 +221,8 @@ void accounting_sta_report(hostapd *hapd, struct sta_info *sta, int stop)
 		goto fail;
 	}
 
-	if (hostapd_read_sta_driver_data(hapd, &data, sta->addr) == 0) {
+	if (hostapd_read_sta_driver_data(hapd->driver.data,
+					 &data, sta->addr) == 0) {
 		if (!radius_msg_add_attr_int32(msg,
 					       RADIUS_ATTR_ACCT_INPUT_PACKETS,
 					       data.rx_packets)) {
