@@ -249,6 +249,13 @@ static int hostapd_config_check(struct hostapd_config *conf)
 		return -1;
 	}
 
+	if (conf->ieee802_1x && conf->minimal_eap &&
+	    (conf->default_wep_key_len || conf->individual_wep_key_len)) {
+		printf("Minimal EAP server cannot be used for dynamic WEP "
+		       "keying.\n");
+		return -1;
+	}
+
 	if (conf->wpa && (conf->wpa_key_mgmt & WPA_KEY_MGMT_PSK) &&
 	    conf->wpa_psk == NULL) {
 		printf("WPA-PSK enabled, but PSK or passphrase is not "
