@@ -21,6 +21,7 @@
 #include "eapol_sm.h"
 #include "eloop.h"
 #include "wpa.h"
+#include "sta_info.h"
 
 /* TODO:
  * implement state machines: Controlled Directions and Key Receive
@@ -58,7 +59,7 @@ static void sm_ ## machine ## _ ## state ## _Enter(struct eapol_state_machine \
 sm->_data.state = machine ## _ ## _state; \
 if (sm->hapd->conf->debug >= HOSTAPD_DEBUG_MINIMAL) \
 	printf("IEEE 802.1X: " MACSTR " " #machine " entering state " #_state \
-		"\n", MAC2STR(sm->sta->addr));
+		"\n", MAC2STR(sm->addr));
 
 #define SM_ENTER(machine, state) sm_ ## machine ## _ ## state ## _Enter(sm)
 
@@ -87,7 +88,7 @@ static void eapol_port_timers_tick(void *eloop_ctx, void *timeout_ctx)
 
 	if (state->hapd->conf->debug >= HOSTAPD_DEBUG_MSGDUMPS)
 		printf("IEEE 802.1X: " MACSTR " Port Timers TICK "
-		       "(timers: %d %d %d %d)\n", MAC2STR(state->sta->addr),
+		       "(timers: %d %d %d %d)\n", MAC2STR(state->addr),
 		       state->aWhile, state->quietWhile, state->reAuthWhen,
 		       state->txWhen);
 
@@ -131,7 +132,7 @@ SM_STATE(AUTH_PAE, DISCONNECTED)
 	if (!from_initialize) {
 		txCannedFail(sm->currentId);
 		sm->currentId++;
-		if (sm->sta->flags & WLAN_STA_PREAUTH)
+		if (sm->flags & EAPOL_SM_PREAUTH)
 			rsn_preauth_finished(sm->hapd, sm->sta, 0);
 	}
 }
@@ -175,9 +176,9 @@ SM_STATE(AUTH_PAE, HELD)
 	sm->auth_pae.eapLogoff = FALSE;
 	sm->currentId++;
 
-	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_IEEE8021X,
+	hostapd_logger(sm->hapd, sm->addr, HOSTAPD_MODULE_IEEE8021X,
 		       HOSTAPD_LEVEL_WARNING, "authentication failed");
-	if (sm->sta->flags & WLAN_STA_PREAUTH)
+	if (sm->flags & EAPOL_SM_PREAUTH)
 		rsn_preauth_finished(sm->hapd, sm->sta, 0);
 }
 
@@ -193,9 +194,9 @@ SM_STATE(AUTH_PAE, AUTHENTICATED)
 	setPortAuthorized();
 	sm->auth_pae.reAuthCount = 0;
 	sm->currentId++;
-	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_IEEE8021X,
+	hostapd_logger(sm->hapd, sm->addr, HOSTAPD_MODULE_IEEE8021X,
 		       HOSTAPD_LEVEL_INFO, "authenticated");
-	if (sm->sta->flags & WLAN_STA_PREAUTH)
+	if (sm->flags & EAPOL_SM_PREAUTH)
 		rsn_preauth_finished(sm->hapd, sm->sta, 1);
 }
 
@@ -395,7 +396,10 @@ SM_STATE(BE_AUTH, FAIL)
 	SM_ENTRY(BE_AUTH, FAIL, be_auth);
 
 	sm->currentId = sm->be_auth.idFromServer;
-	txReq(sm->currentId);
+	if (sm->last_eap_radius == NULL)
+		txCannedFail(sm->currentId);
+	else
+		txReq(sm->currentId);
 	sm->authFail = TRUE;
 }
 
@@ -489,6 +493,7 @@ SM_STATE(REAUTH_TIMER, REAUTHENTICATE)
 	SM_ENTRY(REAUTH_TIMER, REAUTHENTICATE, reauth_timer);
 
 	sm->reAuthenticate = TRUE;
+	wpa_sm_event(sm->hapd, sm->sta, WPA_REAUTH_EAPOL);
 }
 
 
@@ -567,6 +572,10 @@ eapol_sm_alloc(hostapd *hapd, struct sta_info *sta)
 		return NULL;
 	}
 	memset(sm, 0, sizeof(*sm));
+	sm->radius_identifier = -1;
+	memcpy(sm->addr, sta->addr, ETH_ALEN);
+	if (sta->flags & WLAN_STA_PREAUTH)
+		sm->flags |= EAPOL_SM_PREAUTH;
 
 	sm->hapd = hapd;
 	sm->sta = sta;
@@ -619,14 +628,27 @@ void eapol_sm_free(struct eapol_state_machine *sm)
 }
 
 
+static int eapol_sm_sta_entry_alive(struct hostapd_data *hapd, u8 *addr)
+{
+	struct sta_info *sta;
+	sta = ap_get_sta(hapd, addr);
+	if (sta == NULL || sta->eapol_sm == NULL)
+		return 0;
+	return 1;
+}
+
+
 void eapol_sm_step(struct eapol_state_machine *sm)
 {
+	struct hostapd_data *hapd = sm->hapd;
+	u8 addr[6];
 	int prev_auth_pae, prev_be_auth, prev_reauth_timer, prev_auth_key_tx;
 
 	/* FIX: could re-run eapol_sm_step from registered timeout (after
 	 * 0 sec) to make sure that other possible timeouts/events are
 	 * processed */
 
+	memcpy(addr, sm->sta->addr, 6);
 	do {
 		prev_auth_pae = sm->auth_pae.state;
 		prev_be_auth = sm->be_auth.state;
@@ -634,15 +656,24 @@ void eapol_sm_step(struct eapol_state_machine *sm)
 		prev_auth_key_tx = sm->auth_key_tx.state;
 
 		SM_STEP_RUN(AUTH_PAE);
+		if (!eapol_sm_sta_entry_alive(hapd, addr))
+			break;
 		SM_STEP_RUN(BE_AUTH);
+		if (!eapol_sm_sta_entry_alive(hapd, addr))
+			break;
 		SM_STEP_RUN(REAUTH_TIMER);
+		if (!eapol_sm_sta_entry_alive(hapd, addr))
+			break;
 		SM_STEP_RUN(AUTH_KEY_TX);
+		if (!eapol_sm_sta_entry_alive(hapd, addr))
+			break;
 	} while (prev_auth_pae != sm->auth_pae.state ||
 		 prev_be_auth != sm->be_auth.state ||
 		 prev_reauth_timer != sm->reauth_timer.state ||
 		 prev_auth_key_tx != sm->auth_key_tx.state);
 
-	wpa_sm_notify(sm->hapd, sm->sta);
+	if (eapol_sm_sta_entry_alive(hapd, addr))
+		wpa_sm_notify(sm->hapd, sm->sta);
 }
 
 
