@@ -4,8 +4,12 @@
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See README and COPYING for
- * more details.
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
  */
 
 #include <stdlib.h>
@@ -93,6 +97,28 @@ int eloop_register_read_sock(int sock,
 }
 
 
+void eloop_unregister_read_sock(int sock)
+{
+	int i;
+
+	if (eloop.readers == NULL || eloop.reader_count == 0)
+		return;
+
+	for (i = 0; i < eloop.reader_count; i++) {
+		if (eloop.readers[i].sock == sock)
+			break;
+	}
+	if (i == eloop.reader_count)
+		return;
+	if (i != eloop.reader_count - 1) {
+		memmove(&eloop.readers[i], &eloop.readers[i + 1],
+			(eloop.reader_count - i - 1) *
+			sizeof(struct eloop_sock));
+	}
+	eloop.reader_count--;
+}
+
+
 int eloop_register_timeout(unsigned int secs, unsigned int usecs,
 			   void (*handler)(void *eloop_ctx, void *timeout_ctx),
 			   void *eloop_data, void *user_data)
@@ -241,10 +267,10 @@ void eloop_run(void)
 		(eloop.timeout || eloop.reader_count > 0)) {
 		if (eloop.timeout) {
 			gettimeofday(&now, NULL);
-			if (timercmp(&now, &eloop.timeout->time, >=))
-				tv.tv_sec = tv.tv_usec = 0;
-			else
+			if (timercmp(&now, &eloop.timeout->time, <))
 				timersub(&eloop.timeout->time, &now, &tv);
+			else
+				tv.tv_sec = tv.tv_usec = 0;
 #if 0
 			printf("next timeout in %lu.%06lu sec\n",
 			       tv.tv_sec, tv.tv_usec);
@@ -267,7 +293,7 @@ void eloop_run(void)
 			struct eloop_timeout *tmp;
 
 			gettimeofday(&now, NULL);
-			if (timercmp(&now, &eloop.timeout->time, >=)) {
+			if (!timercmp(&now, &eloop.timeout->time, <)) {
 				tmp = eloop.timeout;
 				eloop.timeout = eloop.timeout->next;
 				tmp->handler(tmp->eloop_data,
