@@ -59,7 +59,9 @@ static void hostapd_acl_cache_free(struct hostapd_cached_radius_acl *acl_cache)
 }
 
 
-static int hostapd_acl_cache_get(hostapd *hapd, u8 *addr, u32 *session_timeout)
+static int hostapd_acl_cache_get(struct hostapd_data *hapd, u8 *addr,
+				 u32 *session_timeout,
+				 u32 *acct_interim_interval)
 {
 	struct hostapd_cached_radius_acl *entry;
 	time_t now;
@@ -73,6 +75,7 @@ static int hostapd_acl_cache_get(hostapd *hapd, u8 *addr, u32 *session_timeout)
 				return -1; /* entry has expired */
 			if (entry->accepted == HOSTAPD_ACL_ACCEPT_TIMEOUT)
 				*session_timeout = entry->session_timeout;
+			*acct_interim_interval = entry->acct_interim_interval;
 			return entry->accepted;
 		}
 
@@ -176,6 +179,9 @@ static int hostapd_radius_acl_query(hostapd *hapd, u8 *addr,
 int hostapd_allowed_address(hostapd *hapd, u8 *addr, u8 *msg, size_t len,
 			    u32 *session_timeout, u32 *acct_interim_interval)
 {
+	*session_timeout = 0;
+	*acct_interim_interval = 0;
+
 	if (hostapd_maclist_found(hapd->conf->accept_mac,
 				  hapd->conf->num_accept_mac, addr))
 		return HOSTAPD_ACL_ACCEPT;
@@ -193,7 +199,8 @@ int hostapd_allowed_address(hostapd *hapd, u8 *addr, u8 *msg, size_t len,
 		struct hostapd_acl_query_data *query;
 
 		/* Check whether ACL cache has an entry for this station */
-		int res = hostapd_acl_cache_get(hapd, addr, session_timeout);
+		int res = hostapd_acl_cache_get(hapd, addr, session_timeout,
+						acct_interim_interval);
 		if (res == HOSTAPD_ACL_ACCEPT ||
 		    res == HOSTAPD_ACL_ACCEPT_TIMEOUT)
 			return res;
