@@ -95,8 +95,15 @@ void ieee802_1x_deinit(hostapd *hapd);
 void ieee802_1x_new_auth_session(hostapd *hapd, struct sta_info *sta);
 int ieee802_1x_tx_status(hostapd *hapd, struct sta_info *sta, u8 *buf,
 			 size_t len, int ack);
-int ieee802_1x_send_data(struct hostapd_data *hapd, struct sta_info *sta,
-			 u8 *data, size_t data_len, int encrypt);
-
+u8 * ieee802_1x_get_identity(struct eapol_state_machine *sm, size_t *len);
+u8 * ieee802_1x_get_radius_class(struct eapol_state_machine *sm, size_t *len);
+u8 * ieee802_1x_get_key_crypt(struct eapol_state_machine *sm, size_t *len);
+void ieee802_1x_notify_port_enabled(struct eapol_state_machine *sm,
+				    int enabled);
+void ieee802_1x_notify_port_valid(struct eapol_state_machine *sm,
+				  int valid);
+void ieee802_1x_notify_key_available(struct eapol_state_machine *sm,
+				     int available);
+void ieee802_1x_notify_pre_auth(struct eapol_state_machine *sm, int pre_auth);
 
 #endif /* IEEE802_1X_H */
