@@ -18,6 +18,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
+#include <errno.h>
 
 #include "hostapd.h"
 #include "radius.h"
@@ -42,6 +43,8 @@ static int
 radius_change_server(hostapd *hapd, struct hostapd_radius_server *nserv,
 		     struct hostapd_radius_server *oserv,
 		     int sock, int auth);
+static int radius_client_init_acct(hostapd *hapd);
+static int radius_client_init_auth(hostapd *hapd);
 
 
 static void radius_client_msg_free(struct radius_msg_list *req)
@@ -87,6 +90,26 @@ int radius_client_register(hostapd *hapd, RadiusType msg_type,
 }
 
 
+static void radius_client_handle_send_error(struct hostapd_data *hapd, int s,
+					    RadiusType msg_type)
+{
+	int _errno = errno;
+	perror("send[RADIUS]");
+	if (_errno == ENOTCONN || _errno == EDESTADDRREQ || _errno == EINVAL) {
+		hostapd_logger(hapd, NULL, HOSTAPD_MODULE_RADIUS,
+			       HOSTAPD_LEVEL_INFO,
+			       "Send failed - maybe interface status changed -"
+			       " try to connect again");
+		eloop_unregister_read_sock(s);
+		close(s);
+		if (msg_type == RADIUS_ACCT || msg_type == RADIUS_ACCT_INTERIM)
+			radius_client_init_acct(hapd);
+		else
+			radius_client_init_auth(hapd);
+	}
+}
+
+
 static int radius_client_retransmit(hostapd *hapd,
 				    struct radius_msg_list *entry, time_t now)
 {
@@ -104,7 +127,7 @@ static int radius_client_retransmit(hostapd *hapd,
 		      "\n", entry->msg->hdr->identifier);
 
 	if (send(s, entry->msg->buf, entry->msg->buf_used, 0) < 0)
-		perror("send[RADIUS]");
+		radius_client_handle_send_error(hapd, s, entry->msg_type);
 
 	entry->next_try = now + entry->next_wait;
 	entry->next_wait *= 2;
@@ -333,7 +356,7 @@ int radius_client_send(hostapd *hapd, struct radius_msg *msg,
 
 	res = send(s, msg->buf, msg->buf_used, 0);
 	if (res < 0)
-		perror("send[RADIUS]");
+		radius_client_handle_send_error(hapd, s, msg_type);
 
 	radius_client_list_add(hapd, msg, msg_type, shared_secret,
 			       shared_secret_len, addr);
@@ -662,3 +685,35 @@ void radius_client_deinit(hostapd *hapd)
 	free(hapd->radius);
 	hapd->radius = NULL;
 }
+
+
+void radius_client_flush_auth(struct hostapd_data *hapd, u8 *addr)
+{
+	struct radius_msg_list *entry, *prev, *tmp;
+
+	prev = NULL;
+	entry = hapd->radius->msgs;
+	while (entry) {
+		if (entry->msg_type == RADIUS_AUTH &&
+		    memcmp(entry->addr, addr, ETH_ALEN) == 0) {
+			hostapd_logger(hapd, addr, HOSTAPD_MODULE_RADIUS,
+				       HOSTAPD_LEVEL_DEBUG,
+				       "Removing pending RADIUS authentication"
+				       " message for removed client");
+
+			if (prev)
+				prev->next = entry->next;
+			else
+				hapd->radius->msgs = entry->next;
+
+			tmp = entry;
+			entry = entry->next;
+			radius_client_msg_free(tmp);
+			hapd->radius->num_msgs--;
+			continue;
+		}
+
+		prev = entry;
+		entry = entry->next;
+	}
+}
