@@ -318,14 +318,14 @@ static int wpa_gen_wpa_ie(struct hostapd_data *hapd)
 
 	pos = buf;
 
-	if (hapd->conf->wpa & HOSTAPD_WPA_VERSION_WPA) {
-		res = wpa_write_wpa_ie(hapd, pos, buf + sizeof(buf) - pos);
+	if (hapd->conf->wpa & HOSTAPD_WPA_VERSION_WPA2) {
+		res = wpa_write_rsn_ie(hapd, pos, buf + sizeof(buf) - pos);
 		if (res < 0)
 			return res;
 		pos += res;
 	}
-	if (hapd->conf->wpa & HOSTAPD_WPA_VERSION_WPA2) {
-		res = wpa_write_rsn_ie(hapd, pos, buf + sizeof(buf) - pos);
+	if (hapd->conf->wpa & HOSTAPD_WPA_VERSION_WPA) {
+		res = wpa_write_wpa_ie(hapd, pos, buf + sizeof(buf) - pos);
 		if (res < 0)
 			return res;
 		pos += res;
@@ -347,9 +347,10 @@ static void wpa_sta_disconnect(struct hostapd_data *hapd, struct sta_info *sta)
 	ieee802_11_send_deauth(hapd, sta->addr,
 			       WLAN_REASON_PREV_AUTH_NOT_VALID);
 	sta->flags &= ~(WLAN_STA_AUTH | WLAN_STA_ASSOC | WLAN_STA_AUTHORIZED);
-	remove_sta(hapd, sta);
-	/* FIX: set new timeout for STA entry in hostapd, free 1X/WPA data(?)
-	 */
+	remove_sta(hapd->driver.data, sta->addr);
+	eloop_cancel_timeout(ap_handle_timer, hapd, sta);
+	eloop_register_timeout(0, 0, ap_handle_timer, hapd, sta);
+	sta->timeout_next = STA_REMOVE;
 }
 
 
@@ -452,7 +453,6 @@ static void rsn_preauth_receive(int sock, void *eloop_ctx, void *sock_ctx)
 			return;
 		memset(sta, 0, sizeof(*sta));
 		memcpy(sta->addr, ethhdr->h_source, ETH_ALEN);
-		sta->radius_identifier = -1;
 		sta->flags = WLAN_STA_PREAUTH;
 		sta->next = hapd->sta_list;
 		sta->wpa = WPA_VERSION_WPA2;
@@ -465,7 +465,9 @@ static void rsn_preauth_receive(int sock, void *eloop_ctx, void *sock_ctx)
 			ap_free_sta(hapd, sta);
 			sta = NULL;
 		} else {
+			sta->eapol_sm->radius_identifier = -1;
 			sta->eapol_sm->portValid = TRUE;
+			sta->eapol_sm->flags |= EAPOL_SM_PREAUTH;
 		}
 	}
 	if (sta == NULL)
@@ -554,13 +556,15 @@ static void rsn_preauth_iface_deinit(struct hostapd_data *hapd)
 void rsn_preauth_finished(struct hostapd_data *hapd, struct sta_info *sta,
 			  int success)
 {
+	u8 *key;
+	size_t len;
 	hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_WPA,
 		       HOSTAPD_LEVEL_INFO, "pre-authentication %s",
 		       success ? "succeeded" : "failed");
 
-	if (success && sta->eapol_key_crypt) {
-		pmksa_cache_add(hapd, sta, sta->eapol_key_crypt,
-				dot11RSNAConfigPMKLifetime);
+	key = ieee802_1x_get_key_crypt(sta->eapol_sm, &len);
+	if (success && key) {
+		pmksa_cache_add(hapd, sta, key, dot11RSNAConfigPMKLifetime);
 	}
 
 	ap_free_sta(hapd, sta);
@@ -638,7 +642,8 @@ int wpa_init(struct hostapd_data *hapd)
 		free(tmp);
 	}
 
-	if (hostap_ioctl_prism2param(hapd, PRISM2_PARAM_PRIVACY_INVOKED, 1)) {
+	if (hostap_ioctl_prism2param(hapd->driver.data,
+				     PRISM2_PARAM_PRIVACY_INVOKED, 1)) {
 		printf("Could not set PrivacyInvoked for interface %s\n",
 		       hapd->conf->iface);
 		return -1;
@@ -649,7 +654,8 @@ int wpa_init(struct hostapd_data *hapd)
 		return -1;
 	}
 
-	if (hostapd_set_generic_elem(hapd, hapd->wpa_ie, hapd->wpa_ie_len)) {
+	if (hostapd_set_generic_elem(hapd->driver.data,
+				     hapd->wpa_ie, hapd->wpa_ie_len)) {
 		printf("Failed to configure WPA IE for the kernel driver.\n");
 		return -1;
 	}
@@ -724,12 +730,13 @@ void wpa_deinit(struct hostapd_data *hapd)
 	eloop_cancel_timeout(wpa_rekey_gmk, hapd, NULL);
 	eloop_cancel_timeout(wpa_rekey_gtk, hapd, NULL);
 
-	if (hostap_ioctl_prism2param(hapd, PRISM2_PARAM_PRIVACY_INVOKED, 0)) {
+	if (hostap_ioctl_prism2param(hapd->driver.data,
+				     PRISM2_PARAM_PRIVACY_INVOKED, 0)) {
 		printf("Could not disable PrivacyInvoked for interface %s\n",
 		       hapd->conf->iface);
 	}
 
-	if (hostapd_set_generic_elem(hapd, "", 0)) {
+	if (hostapd_set_generic_elem(hapd->driver.data, "", 0)) {
 		printf("Could not remove generic information element from "
 		       "interface %s\n", hapd->conf->iface);
 	}
@@ -1289,7 +1296,7 @@ void wpa_free_station(struct sta_info *sta)
 		return;
 
 	eloop_cancel_timeout(wpa_send_eapol_timeout, sm->hapd, sta);
-	eloop_cancel_timeout(wpa_sm_call_step, sm->hapd, sta);
+	eloop_cancel_timeout(wpa_sm_call_step, sm->hapd, sta->wpa_sm);
 	free(sm->last_rx_eapol_key);
 	free(sm);
 	sta->wpa_sm = NULL;
@@ -1488,7 +1495,7 @@ void wpa_receive(struct hostapd_data *hapd, struct sta_info *sta,
 				       HOSTAPD_LEVEL_INFO,
 				       "received EAPOL-Key Error Request "
 				       "(STA detected Michael MIC failure)");
-			ieee80211_michael_mic_failure(hapd);
+			ieee80211_michael_mic_failure(hapd, sta->addr, 0);
 			hapd->wpa_auth->dot11RSNAStatsTKIPRemoteMICFailures++;
 			/* Error report is not a request for a new key
 			 * handshake, but since Authenticator may do it, let's
@@ -1758,8 +1765,10 @@ static void wpa_send_eapol(struct hostapd_data *hapd, struct sta_info *sta,
 		if (pad_len)
 			*pos++ = 0xdd;
 
-		hostapd_hexdump("Plaintext EAPOL-Key Key Data",
-				buf, key_data_len);
+		if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MINIMAL)) {
+			hostapd_hexdump("Plaintext EAPOL-Key Key Data",
+					buf, key_data_len);
+		}
 		if (key_info & WPA_KEY_INFO_TYPE_HMAC_SHA1_AES) {
 			aes_wrap(sm->PTK.encr_key, (key_data_len - 8) / 8, buf,
 				 key->key_data);
@@ -1792,7 +1801,10 @@ static void wpa_send_eapol(struct hostapd_data *hapd, struct sta_info *sta,
 				       key->key_mic);
 	}
 
-	ieee802_1x_send_data(hapd, sta, (u8 *) hdr, len, sm->pairwise_set);
+	if (hapd->driver.send_eapol) {
+		hapd->driver.send_eapol(hapd->driver.data, sta->addr,
+					(u8 *) hdr, len, sm->pairwise_set);
+	}
 	free(hdr);
 
 	timeout_ms = pairwise ? dot11RSNAConfigPairwiseUpdateTimeOut :
@@ -1846,6 +1858,7 @@ void wpa_sm_event(struct hostapd_data *hapd, struct sta_info *sta,
 		sm->DeauthenticationRequest = TRUE;
 		break;
 	case WPA_REAUTH:
+	case WPA_REAUTH_EAPOL:
 		sm->ReAuthenticationRequest = TRUE;
 		break;
 	}
@@ -1864,10 +1877,14 @@ void wpa_sm_event(struct hostapd_data *hapd, struct sta_info *sta,
 		sta->eapol_sm->authSuccess = TRUE;
 	}
 
-	sm->pairwise_set = FALSE;
 	sm->PTK_valid = FALSE;
 	memset(&sm->PTK, 0, sizeof(sm->PTK));
-	hostapd_set_encryption(sm->hapd, "none", sm->sta->addr, 0, "", 0);
+
+	if (event != WPA_REAUTH_EAPOL) {
+		sm->pairwise_set = FALSE;
+		hostapd_set_encryption(sm->hapd->driver.data, "none",
+				       sm->sta->addr, 0, "", 0);
+	}
 
 	wpa_sm_step(sm);
 }
@@ -1930,15 +1947,13 @@ SM_STATE(WPA_PTK, INITIALIZE)
 	       * Local AA > Remote AA)) */) {
 		sm->Pair = TRUE;
 	}
-	if (sm->sta->eapol_sm) {
-		sm->sta->eapol_sm->portEnabled = FALSE;
-	}
-	hostapd_set_encryption(sm->hapd, "none", sm->sta->addr, 0, "", 0);
+	ieee802_1x_notify_port_enabled(sm->sta->eapol_sm, 0);
+	hostapd_set_encryption(sm->hapd->driver.data, "none", 
+			       sm->sta->addr, 0, "", 0);
 	sm->pairwise_set = FALSE;
 	sm->PTK_valid = FALSE;
 	memset(&sm->PTK, 0, sizeof(sm->PTK));
-	if (sm->sta->eapol_sm)
-		sm->sta->eapol_sm->portValid = FALSE;
+	ieee802_1x_notify_port_valid(sm->sta->eapol_sm, 0);
 	sm->TimeoutCtr = 0;
 	if (sm->sta->wpa_key_mgmt == WPA_KEY_MGMT_PSK)
 		ieee802_1x_set_sta_authorized(sm->hapd, sm->sta, 0);
@@ -1991,14 +2006,15 @@ SM_STATE(WPA_PTK, AUTHENTICATION2)
 
 SM_STATE(WPA_PTK, INITPMK)
 {
+	u8 *key;
+	size_t len;
 	SM_ENTRY(WPA_PTK, INITPMK, wpa_ptk);
 	if (sm->sta->pmksa) {
 		memcpy(sm->PMK, sm->sta->pmksa->pmk, WPA_PMK_LEN);
-	} else if (sm->sta->eapol_key_crypt) {
-		int len = sm->sta->eapol_key_crypt_len;
+	} else if ((key = ieee802_1x_get_key_crypt(sm->sta->eapol_sm, &len))) {
 		if (len > WPA_PMK_LEN)
 			len = WPA_PMK_LEN;
-		memcpy(sm->PMK, sm->sta->eapol_key_crypt, len);
+		memcpy(sm->PMK, key, len);
 	}
 	sm->sta->req_replay_counter_used = 0;
 	/* IEEE 802.11i/D9.0 does not set keyRun to FALSE, but not doing this
@@ -2080,6 +2096,7 @@ SM_STATE(WPA_PTK, PTKINITNEGOTIATING)
 {
 	u8 rsc[WPA_KEY_RSC_LEN];
 	struct wpa_authenticator *gsm = sm->hapd->wpa_auth;
+	u8 *wpa_ie;
 	int wpa_ie_len;
 
 	SM_ENTRY(WPA_PTK, PTKINITNEGOTIATING, wpa_ptk);
@@ -2087,13 +2104,15 @@ SM_STATE(WPA_PTK, PTKINITNEGOTIATING)
 	/* Send EAPOL(1, 1, 1, Pair, P, RSC, ANonce, MIC(PTK), RSNIE, GTK[GN])
 	 */
 	memset(rsc, 0, WPA_KEY_RSC_LEN);
-	hostapd_get_seqnum(sm->hapd, NULL, gsm->GN, rsc);
+	hostapd_get_seqnum(sm->hapd->driver.data, NULL, gsm->GN, rsc);
+	wpa_ie = sm->hapd->wpa_ie;
 	wpa_ie_len = sm->hapd->wpa_ie_len;
 	if (sm->sta->wpa == WPA_VERSION_WPA &&
 	    (sm->hapd->conf->wpa & HOSTAPD_WPA_VERSION_WPA2) &&
-	    wpa_ie_len > sm->hapd->wpa_ie[1] + 2) {
+	    wpa_ie_len > wpa_ie[1] + 2 && wpa_ie[0] == WLAN_EID_RSN) {
 		/* WPA-only STA, remove RSN IE */
-		wpa_ie_len = sm->hapd->wpa_ie[1] + 2;
+		wpa_ie = wpa_ie + wpa_ie[1] + 2;
+		wpa_ie_len = wpa_ie[1] + 2;
 	}
 	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_WPA,
 		       HOSTAPD_LEVEL_DEBUG,
@@ -2101,7 +2120,7 @@ SM_STATE(WPA_PTK, PTKINITNEGOTIATING)
 	wpa_send_eapol(sm->hapd, sm->sta,
 		       sm->sta->wpa == WPA_VERSION_WPA2 ? 1 : 0,
 		       1, 1, 1, 1, rsc, sm->ANonce,
-		       sm->hapd->wpa_ie, wpa_ie_len,
+		       wpa_ie, wpa_ie_len,
 		       gsm->GTK[gsm->GN - 1], gsm->GTK_len, gsm->GN);
 	sm->TimeoutCtr++;
 }
@@ -2121,7 +2140,8 @@ SM_STATE(WPA_PTK, PTKINITDONE)
 			alg = "CCMP";
 			klen = 16;
 		}
-		if (hostapd_set_encryption(sm->hapd, alg, sm->sta->addr, 0,
+		if (hostapd_set_encryption(sm->hapd->driver.data, 
+					   alg, sm->sta->addr, 0,
 					   sm->PTK.tk1, klen)) {
 			wpa_sta_disconnect(sm->hapd, sm->sta);
 			return;
@@ -2136,19 +2156,21 @@ SM_STATE(WPA_PTK, PTKINITDONE)
 	if (0 /* IBSS == TRUE */) {
 		sm->keycount++;
 		if (sm->keycount == 2) {
-			if (sm->sta->eapol_sm)
-				sm->sta->eapol_sm->portValid = TRUE;
+			ieee802_1x_notify_port_valid(sm->sta->eapol_sm, 1);
 		}
 	} else {
-		if (sm->sta->eapol_sm)
-			sm->sta->eapol_sm->portValid = TRUE;
+		ieee802_1x_notify_port_valid(sm->sta->eapol_sm, 1);
 	}
-	if (sm->sta->eapol_sm)
+	if (sm->sta->eapol_sm) {
+		sm->sta->eapol_sm->keyAvailable = FALSE;
 		sm->sta->eapol_sm->keyDone = TRUE;
+	}
 	if (sm->sta->wpa == WPA_VERSION_WPA)
 		sm->PInitAKeys = TRUE;
 	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_WPA,
-		       HOSTAPD_LEVEL_INFO, "pairwise key handshake completed");
+		       HOSTAPD_LEVEL_INFO, "pairwise key handshake completed "
+		       "(%s)",
+		       sm->sta->wpa == WPA_VERSION_WPA ? "WPA" : "RSN");
 }
 
 
@@ -2256,7 +2278,8 @@ SM_STATE(WPA_PTK_GROUP, REKEYNEGOTIATING)
 	sm->TimeoutEvt = FALSE;
 	/* Send EAPOL(1, 1, 1, !Pair, G, RSC, GNonce, MIC(PTK), GTK[GN]) */
 	memset(rsc, 0, WPA_KEY_RSC_LEN);
-	hostapd_get_seqnum(sm->hapd, NULL, gsm->GN, rsc);
+	if (gsm->wpa_group_state == WPA_GROUP_SETKEYSDONE)
+		hostapd_get_seqnum(sm->hapd->driver.data, NULL, gsm->GN, rsc);
 	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_WPA,
 		       HOSTAPD_LEVEL_DEBUG,
 		       "sending 1/2 msg of Group Key Handshake");
@@ -2276,7 +2299,9 @@ SM_STATE(WPA_PTK_GROUP, REKEYESTABLISHED)
 	sm->GTimeoutCtr = 0;
 	/* FIX: MLME.SetProtection.Request(TA, Tx_Rx) */
 	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_WPA,
-		       HOSTAPD_LEVEL_INFO, "group key handshake completed");
+		       HOSTAPD_LEVEL_INFO, "group key handshake completed "
+		       "(%s)",
+		       sm->sta->wpa == WPA_VERSION_WPA ? "WPA" : "RSN");
 }
 
 
@@ -2329,7 +2354,7 @@ static void wpa_group_gtk_init(struct hostapd_data *hapd)
 	/* GTK[0..N] = 0 */
 	memset(sm->GTK, 0, sizeof(sm->GTK));
 	sm->GN = 1;
-	sm->GN = 2;
+	sm->GM = 2;
 	/* GTK[GN] = CalcGTK() */
 	/* FIX: is this the correct way of getting GNonce? */
 	memcpy(sm->GNonce, sm->Counter, WPA_NONCE_LEN);
@@ -2379,7 +2404,8 @@ static void wpa_group_setkeysdone(struct hostapd_data *hapd)
 		      "entering state SETKEYSDONE\n");
 	sm->changed = TRUE;
 	sm->wpa_group_state = WPA_GROUP_SETKEYSDONE;
-	hostapd_set_encryption(hapd, wpa_alg_txt(hapd->conf->wpa_group),
+	hostapd_set_encryption(hapd->driver.data,
+			       wpa_alg_txt(hapd->conf->wpa_group),
 			       NULL, sm->GN, sm->GTK[sm->GN - 1],
 			       sm->GTK_len);
 }
@@ -2406,15 +2432,37 @@ static void wpa_group_sm_step(struct hostapd_data *hapd)
 }
 
 
+static int wpa_sm_sta_entry_alive(struct hostapd_data *hapd, u8 *addr)
+{
+	struct sta_info *sta;
+	sta = ap_get_sta(hapd, addr);
+	if (sta == NULL || sta->wpa_sm == NULL)
+		return 0;
+	return 1;
+}
+
+
 static void wpa_sm_step(struct wpa_state_machine *sm)
 {
+	struct hostapd_data *hapd = sm->hapd;
+	u8 addr[6];
+	if (sm == NULL || sm->sta == NULL || sm->sta->wpa_sm == NULL)
+		return;
+
+	memcpy(addr, sm->sta->addr, 6);
 	do {
 		sm->changed = FALSE;
 		sm->hapd->wpa_auth->changed = FALSE;
 
 		SM_STEP_RUN(WPA_PTK);
+		if (!wpa_sm_sta_entry_alive(hapd, addr))
+			break;
 		SM_STEP_RUN(WPA_PTK_GROUP);
+		if (!wpa_sm_sta_entry_alive(hapd, addr))
+			break;
 		wpa_group_sm_step(sm->hapd);
+		if (!wpa_sm_sta_entry_alive(hapd, addr))
+			break;
 	} while (sm->changed || sm->hapd->wpa_auth->changed);
 }
 
