@@ -162,7 +162,8 @@ void wpa_free_station(struct sta_info *sta);
 void wpa_receive(struct hostapd_data *hapd, struct sta_info *sta,
 		 u8 *data, size_t data_len);
 typedef enum {
-	WPA_AUTH, WPA_ASSOC, WPA_DISASSOC, WPA_DEAUTH, WPA_REAUTH
+	WPA_AUTH, WPA_ASSOC, WPA_DISASSOC, WPA_DEAUTH, WPA_REAUTH,
+	WPA_REAUTH_EAPOL
 } wpa_event;
 void wpa_sm_event(struct hostapd_data *hapd, struct sta_info *sta,
 		  wpa_event event);
